/*
	BeOS VNC Server by Mikael Eiman (mikael@eiman.tv)
	Copyright 2001
	
	The following rules apply to the use of this code:
	
	* The code may be freely distributed, as long as this comment isn't removed or edited
	* The code may not be used for commercial purposes without the explicit permission of the author
	* If you fix anything, send a copy of the corrected code to the author!
*/

#include <Application.h>
#include <Screen.h>
#include <stdio.h>
#include <Bitmap.h>
#include <DirectWindow.h>
#include <Locker.h>
#include <memory.h>
#include <NetEndpoint.h>
#include <ByteOrder.h>
#include <WindowScreen.h>
#include <Errors.h>
#include <View.h>
#include <List.h>
#include <Roster.h>
#include <string.h>
#include <stdlib.h>

// Get frames of all windows visible on current workspace
// This function seems to lock on certain apps when they do certain things.
// It appears that when they do synchronous mouse trackings, things stop.
// This shouldn't happen, since I've got timeouts on all the messaging, but it does anyway!
void update_window_frames( BList * rects )
{
	//#define DEBUG_WINDOW_LIST
	
	#ifdef DEBUG_WINDOW_LIST
	printf("Updating window list\n");
	#endif

	// clear list
	for ( int c=0; c<rects->CountItems(); c++ )
		delete (BRect*)rects->ItemAt(c);
	rects->MakeEmpty();
	
	app_info my_app;
	be_app->GetAppInfo( &my_app );
	
	BList app_list;
	
	#ifdef DEBUG_WINDOW_LIST
	printf("  Getting app list\n");
	#endif
	be_roster->GetAppList( &app_list );
	
	for ( int c=0; c<app_list.CountItems(); c++ )
	{
		team_id team = (team_id)app_list.ItemAt(c);
		
		if ( team == my_app.team )
			continue;
		
		#ifdef DEBUG_WINDOW_LIST
		printf("  Gettings info on app\n");
		#endif
		app_info info;
		be_roster->GetRunningAppInfo( team, &info );
		
		BMessenger msgr(NULL,team);
		BMessage reply;
		
		// we should ask the apps how many windows they have, but they're lying bastards!
/*		int32 num_windows=0;
		BMessage count_request;
		count_request.what = B_COUNT_PROPERTIES;
		count_request.AddSpecifier("Window");
		msgr.SendMessage( &count_request, &reply );
		
		if ( reply.FindInt32("result", &num_windows) != B_OK )
		{
			printf("app %ld %s has no windows\n", team, info.signature );
			continue;
		}
*/		
		#define MAX_DELAY 10000
		
		#ifdef DEBUG_WINDOW_LIST
		printf("  Listing windows\n");
		#endif
		for ( int curr_win=0; /*curr_win<num_windows*/; curr_win++ )
		{
			BMessage hidden_request;
			hidden_request.what = B_GET_PROPERTY;
			hidden_request.AddSpecifier("Hidden");
			hidden_request.AddSpecifier("Window", curr_win);
			reply.MakeEmpty();
			if ( msgr.SendMessage(&hidden_request,&reply,MAX_DELAY) != B_OK )
				break;
			
			bool hidden=true;
			if ( reply.FindBool("result", &hidden) != B_OK || hidden )
				break;
			
			BMessage minimized_request;
			minimized_request.what = B_GET_PROPERTY;
			minimized_request.AddSpecifier("Minimize");
			minimized_request.AddSpecifier("Window", curr_win);
			reply.MakeEmpty();
			if ( msgr.SendMessage(&minimized_request,&reply,MAX_DELAY) != B_OK )
				break;
			
			bool minimized=true;
			if ( reply.FindBool("result", &minimized) != B_OK || minimized )
				break;
			
			BMessage workspaces_request;
			workspaces_request.what = B_GET_PROPERTY;
			workspaces_request.AddSpecifier("Workspaces");
			workspaces_request.AddSpecifier("Window", curr_win);
			reply.MakeEmpty();
			if ( msgr.SendMessage(&workspaces_request,&reply,MAX_DELAY) != B_OK )
				break;
			
			int32 workspaces=0;
			if ( reply.FindInt32("result", &workspaces) != B_OK || (workspaces & (1 << current_workspace())) == 0 )
				break;
			
			if ( strcmp(info.signature, "application/x-vnd.Be-TRAK") == 0 )
			{ // skip desktop if tracker
				BMessage title_request;
				title_request.what = B_GET_PROPERTY;
				title_request.AddSpecifier("Title");
				title_request.AddSpecifier("Window", curr_win);
				reply.MakeEmpty();
				if ( msgr.SendMessage(&title_request,&reply,MAX_DELAY) != B_OK )
					break;
				
				const char * title;
				
				if ( reply.FindString("result", &title) != B_OK )
					break;
				
				if ( strcmp(title,"Desktop") == 0 )
					continue;
			}
			
			BMessage frame_request;
			frame_request.what = B_GET_PROPERTY;
			frame_request.AddSpecifier("Frame");
			frame_request.AddSpecifier("Window", curr_win);
			reply.MakeEmpty();
			if ( msgr.SendMessage(&frame_request,&reply,MAX_DELAY) != B_OK )
				break;
			
			BRect frame;
			if ( reply.FindRect("result", &frame) != B_OK )
				break;
			
			// modify so it works better with DoScreenshot
			frame.top -= 24; // for title
			frame.InsetBy(-4,-4); // for borders
			frame.right += 1;
			frame.bottom += 1;
			// Add to list
			rects->AddItem( new BRect(frame) );
		}
		#ifdef DEBUG_WINDOW_LIST
		printf("  Done updating window list\n");
		#endif
	}
	#ifdef DEBUG_WINDOW_LIST
	printf("Done updating window list\n");
	#endif
}

// used by client_thread() and window_list_updater()
typedef struct window_list_t {
	BLocker		lock; // lock before handling list
	BList		* list; // list of BRect*
	sem_id		update_list; // release to request a window list update
	thread_id	updater_thread; // id of updater thread
};

/*
	Updates the window list.
*/
int32 window_list_updater( void * _data )
{
	window_list_t * data = (window_list_t*)_data;
	
	while ( true )
	{
		// wait for update request
		if ( acquire_sem( data->update_list ) != B_OK )
			return 0;
		
		// get new list
		BList * new_list = new BList;
		update_window_frames( new_list );
		
		// replace old list
		data->lock.Lock();
		for ( int c=0; c<data->list->CountItems(); c++ )
			delete (BRect*)data->list->ItemAt(c);
		delete data->list;
		
		data->list = new_list;
		data->lock.Unlock();
	}
}

// used by update_frame_info
typedef struct frame_info {
	int width;
	int height;
	int bytes_per_pixel;
	
	uint8 * prev_buffer;
	uint8 * next_buffer;
	
	int		tiles_per_line;
	int		tile_lines;
	bool	* tiles;
};


/*
	compare two frames
*/
void update_frame_info( frame_info * info )
{
	uint32 bytes_per_line = info->width * info->bytes_per_pixel;
	
	for ( int tile_y=0; tile_y<info->tile_lines; tile_y++ )
	{
		for ( int tile_x=0; tile_x<info->tiles_per_line; tile_x++ )
		{
			uint32 top_left_offset = (tile_y*16) * info->width * info->bytes_per_pixel;
			top_left_offset += tile_x * 16 * info->bytes_per_pixel;
			
			int32 to_compare = 16;
			if ( (tile_x+1) * 16 > info->width )
				to_compare = info->width % 16;
			to_compare *= info->bytes_per_pixel;
			
			int32 lines = 16;
			if ( (tile_y+1) * 16 > info->height )
			{
				lines = info->height%16;
			}
			
			for ( int l=0; l<lines; l++ )
			{
				uint32 offset = top_left_offset + l*bytes_per_line;
				if ( memcmp(&info->prev_buffer[offset], &info->next_buffer[offset], to_compare) )
				{
					info->tiles[ tile_y*info->tiles_per_line + tile_x] = true;
					break;
				}
			}
		}
	}
}

// Data types

typedef uint8 CARD8;
typedef uint16 CARD16;
typedef uint32 CARD32;

// Init messages

enum {
	AUTHEN_CONNECTION_FAILED = 0,
	AUTHEN_NO_AUTHENTICATION = 1,
	AUTHEN_VNC_AUTHENTICATION = 2,
	
	AUTHEN_OK = 0,
	AUTHEN_FAILED = 1,
	AUTHEN_TOO_MANY = 2
};

typedef struct authentication_t
{
	CARD32	scheme;
};

typedef struct client_init_t
{
	CARD8	shared;
};

typedef struct PIXEL_FORMAT
{
	CARD8	bits_per_pixel;
	CARD8	depth;
	CARD8	big_endian;
	CARD8	true_color;
	CARD16	red_max;
	CARD16	green_max;
	CARD16	blue_max;
	CARD8	red_shift;
	CARD8	green_shift;
	CARD8	blue_shift;

	CARD8	padding[3];
};

typedef struct server_init_t
{
	CARD16			framebuffer_width;
	CARD16			framebuffer_height;
	PIXEL_FORMAT	server_pixel_format;
	CARD32			name_length;
	// then the name data
};

// client to server messages

typedef struct set_pixel_format_t
{
	CARD8			message_type; // == 0
	CARD8			pad[3];
	
	PIXEL_FORMAT	pixel_format;
};

typedef struct set_encodings_t
{
	CARD8		message_type; // == 2
	CARD8		pad;
	
	CARD16		number_of_encodings;
	
	// followed by number_of_encodings * CARD32
};

enum encoding_t {
	ENCODING_RAW = 0,
	ENCODING_COPY_RECTANGLE = 1,
	ENCODING_RRE = 2,
	ENCODING_CoRRE = 4,
	ENCODING_HEXTILE = 5
};

typedef struct framebuffer_update_request_t
{
	CARD8	message_type; // == 3
	CARD8	incremential;
	CARD16	x;
	CARD16	y;
	CARD16	width;
	CARD16	height;
};

typedef struct key_event_t
{
	CARD8	message_type; // = 4
	CARD8	down;
	CARD8	pad[2];
	CARD32	key; // see <X11/keysymdef.h> (as if I've got that one lying around ;)
};

typedef struct pointer_event_t
{
	CARD8	message_type; // == 5
	CARD8	buttons;
	CARD16	x;
	CARD16	y;
};

typedef struct client_cut_text_t
{
	CARD8	message_type; // == 6
	CARD8	pad[3];
	CARD32	length;
	// followed by CARD8 * length
};

// server to client messages

typedef struct framebuffer_update_t
{
	CARD8	message_type; // == 0
	CARD8	pad;
	CARD16	num_rect;
	// followed by num_rect * rectangle_t
};

typedef struct rectangle_t
{
	CARD16	x;
	CARD16	y;
	CARD16	width;
	CARD16	height;
	CARD32	encoding;
	// followed by data as defined by encoding
};

typedef struct copy_rectangle_t
{
	CARD16	x;
	CARD16	y;
};

typedef struct hextile_t
{
	CARD8	subencoding;
	/*	bit flags for subencoding
		1	raw
		2	background specified
		4	foreground specified
		8	are there any subrects
		16	subrects coloured
	*/
};

typedef struct client_data_t
{
	BNetEndpoint	* endpoint;
	
	int		frames_between_full_update;
	bool	compress_bitdepth;
};

int32 g_total_data_read = 0; // used for stats/info
// read data to endpoint
// retry if interrupted (bad net_server, BAD net_server!)
int endpoint_read( BNetEndpoint * endpoint, void * _buffer, int32 len )
{
	char * buffer = (char*)_buffer;
	
	int32 curr = 0;
	
	while ( curr < len )
	{
		int32 res = endpoint->Receive(&buffer[curr], len-curr);
		
		if ( res == 0 )
		{ // disconnected?
			return B_ERROR;
		}
		
		if ( res > 0 )
		{ // ok
			curr += res;
		}
		
		if ( res < 0 )
		{ // error
			switch ( endpoint->Error() )
			{
				case EINTR: // snooze a bit if interrupted, then continue
					snooze(10000);
					break;
				default:
					printf("Network error: %s\n", endpoint->ErrorStr() );
					return B_ERROR;
			}
		}
	}
	
	g_total_data_read += len;
	return len;
}

int32 g_total_data_sent = 0; // used for stats/info
// send data to endpoint
// retry if interrupted (bad net_server, BAD net_server!)
int endpoint_write( BNetEndpoint * endpoint, const void * _buffer, int32 len )
{
	const char * buffer = (const char*)_buffer;
	
	int32 curr = 0;
	
	while ( curr < len )
	{
		int32 res = endpoint->Send(&buffer[curr], len-curr);
		
		if ( res == 0 )
		{ // disconnected?
			return B_ERROR;
		}
		
		if ( res > 0 )
		{ // ok!
			curr += res;
		}
		
		if ( res < 0 )
		{ // error
			switch ( endpoint->Error() )
			{
				case EINTR: // if we're interrupted, sleep a while and continue
					snooze(10000);
					break;
				default:
					printf("Network error: %s\n", endpoint->ErrorStr() );
					return B_ERROR;
			}
		}
	}
	
	g_total_data_sent += len;
	return len;
}

/*
	Provides access to the frame buffer. Too bad it's so SLOW to read from it.. :P
*/
class FBAWindow : public BDirectWindow
{
	private:
		uint8			* m_bits;
		int32			m_bytes_per_row;
		color_space		m_color_space;
		bool			m_connected;
		
	public:
		FBAWindow()
		:	BDirectWindow(
				BRect(0,0,1,1), "",
				B_NO_BORDER_WINDOW_LOOK,
				B_FLOATING_ALL_WINDOW_FEEL,
				B_NOT_MOVABLE|B_NOT_ZOOMABLE|B_NOT_RESIZABLE
			),
			m_connected( false )
		{
			BView * view = new BView(
				Bounds(), "", B_FOLLOW_NONE, 0
			);		
			view->SetViewColor( B_TRANSPARENT_COLOR );
			AddChild( view );
			
			SetSizeLimits(1,1,1,1);
			ResizeTo(1,1);
			
			SetWorkspaces( B_ALL_WORKSPACES ); // All workspaces!
			
			if ( !SupportsWindowMode() )
				printf("FBAWindow: No support for windowed mode\n");
			
			Show();
		};
		
		virtual ~FBAWindow()
		{
		};
		
		virtual void DirectConnected( direct_buffer_info * info )
		{
			switch ( info->buffer_state & B_DIRECT_MODE_MASK )
			{
				case B_DIRECT_START:
				{ // connected
					printf("FBAWindow: Connected\n");
					m_connected = true;
				}	// fall through!
				case B_DIRECT_MODIFY:
				{ // something changed
					printf("FBAWindow: Updating direct access info\n");
					m_bits = (uint8*)info->bits;
					m_bytes_per_row = info->bytes_per_row;
					m_color_space = info->pixel_format;
				}	break;
				
				case B_DIRECT_STOP:
				{ // disconnected
					printf("FBAWindow: Disconnected\n");
					m_connected = false;
				}	break;
			};
		};
		
		
		/*
			Do a screenshot, converting colorspaces as needed
		*/
		bool DoScreenshot( uint8 * buffer, BRect rect, PIXEL_FORMAT * pixel_format )
		{
			BLocker locker(this);
			
			if ( !m_connected )
				return false;
			
			BScreen screen( this );
			
			BRect frame = screen.Frame();
			
			int32 bytes_per_pixel = 4;
			if ( screen.ColorSpace() == B_RGB16 )
				bytes_per_pixel = 2;
			
			int32 width = (int32)frame.Width() + 1;
			//int32 height = (int32)frame.Height() + 1;
			
			for ( int y=(int)rect.top; y<(int)rect.bottom; y++ )
			{
				int32 * fb_line = (int32*)&m_bits[m_bytes_per_row * y + (int32)rect.left*bytes_per_pixel];
				int32 * local_line = (int32*)&buffer[(pixel_format->bits_per_pixel/8)*width * y + (int32)rect.left*(pixel_format->bits_per_pixel/8)];
				
				// for each pixel, convert to suiting format, or memcpy if already the right one
				if ( bytes_per_pixel == pixel_format->bits_per_pixel/8 )
				{ // at correct bpp
					memcpy(local_line, fb_line, (int32)rect.Width()*bytes_per_pixel);
				} else
				{ // convert
					if ( bytes_per_pixel == 2 )
					{ // 16bpp
						uint16 * framebuffer = (uint16*)fb_line;
						uint8 * target = (uint8*)local_line;
						
						uint8 red_shift = pixel_format->red_shift;
						uint8 green_shift = pixel_format->green_shift;
						uint8 blue_shift = pixel_format->blue_shift;

						for ( int x=(int)rect.left; x<(int)rect.right; x++ )
						{
							uint16 pixel = *framebuffer;
							
							*target = ( (((pixel >> 11) & 31) >> 2) << red_shift ) | 
							          ( (((pixel >> 5) & 63) >> 3) << green_shift ) | 
							          ( (((pixel >> 0) & 31) >> 2) << blue_shift );
							target++;
							framebuffer++;
						}
					} else
					{ // 32bpp
						uint32 * framebuffer = (uint32*)fb_line;
						uint8 * target = (uint8*)local_line;
						
						uint8 red_shift = pixel_format->red_shift;
						uint8 green_shift = pixel_format->green_shift;
						uint8 blue_shift = pixel_format->blue_shift;
						
						for ( int x=(int)rect.left; x<(int)rect.right; x++ )
						{
							uint32 pixel = *framebuffer;
							
							*target = ( (((pixel >> 16) & 255) >> 2) << red_shift ) | 
							          ( (((pixel >> 8) & 255) >> 3) << green_shift ) | 
							          ( (((pixel >> 0) & 255) >> 2) << blue_shift );
							target++;
							framebuffer++;
						}
					}
				}
			}
			
			return true;
		};
		
		
		/*
			Return current mouse position
		*/
		BPoint GetMousePos()
		{
			BPoint point;
			uint32 buttons;
			ChildAt(0)->GetMouse(&point,&buttons);
			
			return point;
		};
};

// Global variable, points to our framebuffer access window.
// used by client_thread to perform screenshots
FBAWindow * g_fba_window = NULL;


// mouse pointer data, since we can't at get the actual pointer
class pixel_t {
	public:
		bool	solid;
		uint32	data;
};
#define W 0xffffffff
#define B 0x00000000
#define POINTER_WIDTH 16
#define POINTER_HEIGHT 16
const pixel_t mouse_pointer[POINTER_HEIGHT][POINTER_WIDTH] = {
	{ {1,B},{1,B},{1,B},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0} },
	{ {1,B},{1,W},{1,W},{1,B},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0} },
	{ {1,B},{1,W},{1,W},{1,B},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0} },
	{ {1,B},{1,W},{1,W},{1,W},{1,B},{1,B},{1,B},{1,B},{1,B},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0} },
	{ {0,0},{1,B},{1,W},{1,W},{1,B},{1,W},{1,W},{1,B},{1,W},{1,B},{1,B},{1,B},{0,0},{0,0},{0,0},{0,0} },
	{ {0,0},{1,B},{1,W},{1,W},{1,W},{1,B},{1,W},{1,W},{1,B},{1,W},{1,B},{1,W},{1,B},{0,0},{0,0},{0,0} },
	{ {0,0},{0,0},{1,B},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,B},{0,0},{0,0},{0,0} },
	{ {0,0},{0,0},{1,B},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,B},{0,0},{0,0},{0,0} },
	{ {0,0},{1,B},{1,B},{1,B},{1,B},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,B},{0,0},{0,0},{0,0} },
	{ {1,B},{1,W},{1,W},{1,B},{1,B},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,B},{0,0},{0,0},{0,0} },
	{ {1,B},{1,W},{1,W},{1,W},{1,W},{1,B},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,B},{0,0},{0,0},{0,0} },
	{ {1,B},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,B},{0,0},{0,0},{0,0} },
	{ {0,0},{1,B},{1,B},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,B},{0,0},{0,0},{0,0} },
	{ {0,0},{0,0},{0,0},{1,B},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{1,B},{0,0},{0,0},{0,0} },
	{ {0,0},{0,0},{0,0},{0,0},{1,B},{1,B},{1,W},{1,W},{1,W},{1,W},{1,W},{1,W},{0,0},{0,0},{0,0},{0,0} },
	{ {0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{1,B},{1,W},{1,W},{1,W},{1,W},{0,0},{0,0},{0,0},{0,0},{0,0} }
};


/*
	Find a given key in a given keymap, return true if found.
	Adds a bunhc of info to message if the key is found.
*/
bool find_key_in_keymap( int32 key, int32 * map, char * chars, BMessage * msg, int32 modif )
{
	char string[4] = {0,0,0,0};
	
	for ( int c=0; c<128; c++ )
	{
		if ( chars[ map[c] + 1 ] == key )
		{
			msg->AddInt32("key", c);
			msg->AddInt32("modifiers", 0);
			memcpy( string, &chars[map[c]+1], chars[map[c]] );
			msg->AddString("bytes", string);
			for ( int i=0; string[i]; i++ )
				msg->AddInt8("byte", string[i]);
			msg->AddInt32("modifiers",modif);
			return true;
		}
	}
	
	return false;
}

/*
	Auto-resizing buffer, used by client_thread to buffer network writes.
*/
class SmartBuffer
{
	private:
		int32	m_size;
		int32	m_used;
		char	* m_buffer;
		
	public:
		SmartBuffer( int32 initial_size )
		:	m_size( initial_size ),
			m_used(0)
		{
			m_buffer = (char*)malloc(m_size);
		};
		
		~SmartBuffer()
		{
			free( m_buffer );
		};
		
		int32 Size()
		{ // size of USED buffer space!
			return m_used;
		};
		
		void Add( const void * data, int32 size )
		{
			if ( m_used + size > m_size )
			{ // buffer too small, resize to fit
				int32 new_size = m_size;
				while ( new_size < m_used + size )
					new_size *= 2;
				
				m_buffer = (char*)realloc( m_buffer, new_size );
			}
			
			memcpy( &m_buffer[m_used], data, size );
			
			m_used += size;
		};
		
		void * Buffer()
		{
			return (void*)m_buffer;
		};
};


/*
	Handles a client session. Should be broken up into a bunch of functions, but I haven't gotten around to it yet.
*/
int32 client_thread( void * _data )
{
	BScreen screen;
	int32 screen_width = (int32)screen.Frame().Width() + 1;
	int32 screen_height = (int32)screen.Frame().Height() + 1;
	
	printf("Screen dimensions %ldx%ld\n", screen_width, screen_height);
	
	client_data_t	* data = (client_data_t*)_data;
	
	BNetEndpoint * endpoint = data->endpoint;
	
	printf("Client connect\n");
	
	{ // negotiate version
		// send preferred version to client
		endpoint_write(endpoint, "RFB 003.003\n", 12 );
		
		// receive selected version from client
		char version[13];
		endpoint_read(endpoint, version, 12);
		version[12] = 0;
		printf("Version selected: %s", version);
	}
		
	{ // send authentication request
		authentication_t authen;
		authen.scheme = B_HOST_TO_BENDIAN_INT32( AUTHEN_VNC_AUTHENTICATION );
		CARD8 challenge[16];
		CARD8 response[16];
	
		endpoint_write( endpoint, &authen, sizeof(authen) );
		endpoint_write( endpoint, challenge, sizeof(challenge) );
		endpoint_read( endpoint, response, sizeof(response) );
	
		CARD32 authen_result = AUTHEN_OK;
	
		endpoint_write( endpoint, &authen_result, sizeof(authen_result) );
	}
	
	{ // receive client init
		client_init_t client_init;
		endpoint_read( endpoint, &client_init, sizeof(client_init) );
		printf("Client shared: %d\n", (int)client_init.shared );
	}
	
	// send server init message
	const char * connection_name = "BeOS VNC";
	server_init_t server_init;
	server_init.framebuffer_width = B_HOST_TO_BENDIAN_INT16( screen_width );
	server_init.framebuffer_height = B_HOST_TO_BENDIAN_INT16( screen_height );
	
	// pixel format
	if ( data->compress_bitdepth )
	{
		server_init.server_pixel_format.bits_per_pixel = 8;
		server_init.server_pixel_format.depth = 8;
		server_init.server_pixel_format.big_endian = false;
		server_init.server_pixel_format.true_color = true;
		server_init.server_pixel_format.red_max = B_HOST_TO_BENDIAN_INT16( 7 );
		server_init.server_pixel_format.green_max = B_HOST_TO_BENDIAN_INT16( 7 );
		server_init.server_pixel_format.blue_max = B_HOST_TO_BENDIAN_INT16( 3 );
		server_init.server_pixel_format.red_shift = 5;
		server_init.server_pixel_format.green_shift = 2;
		server_init.server_pixel_format.blue_shift = 0;
	} else
	{
		switch ( screen.ColorSpace() )
		{
			case B_RGB32:
				server_init.server_pixel_format.bits_per_pixel = 32;
				server_init.server_pixel_format.depth = 24; // 32?
				server_init.server_pixel_format.big_endian = false;
				server_init.server_pixel_format.true_color = true;
				server_init.server_pixel_format.red_max = B_HOST_TO_BENDIAN_INT16( 255 );
				server_init.server_pixel_format.green_max = B_HOST_TO_BENDIAN_INT16( 255 );
				server_init.server_pixel_format.blue_max = B_HOST_TO_BENDIAN_INT16( 255 );
				server_init.server_pixel_format.red_shift = 16;
				server_init.server_pixel_format.green_shift = 8;
				server_init.server_pixel_format.blue_shift = 0;
				break;
			case B_RGB16:
				server_init.server_pixel_format.bits_per_pixel = 16;
				server_init.server_pixel_format.depth = 16;
				server_init.server_pixel_format.big_endian = false;
				server_init.server_pixel_format.true_color = true;
				server_init.server_pixel_format.red_max = B_HOST_TO_BENDIAN_INT16( 31 );
				server_init.server_pixel_format.green_max = B_HOST_TO_BENDIAN_INT16( 63 );
				server_init.server_pixel_format.blue_max = B_HOST_TO_BENDIAN_INT16( 31 );
				server_init.server_pixel_format.red_shift = 11;
				server_init.server_pixel_format.green_shift = 5;
				server_init.server_pixel_format.blue_shift = 0;
				break;
			default:
				printf("Other colorspace %d\n", screen.ColorSpace() );
				break;
		}
	}
	server_init.name_length = B_HOST_TO_BENDIAN_INT32( strlen(connection_name) );
	// transmit struct
	endpoint_write( endpoint, &server_init, sizeof(server_init) );
	endpoint_write( endpoint, connection_name, strlen(connection_name) );
	
	// previous buffer pointer
	uint8 * last_buffer = NULL;
	// pixel format to be used
	PIXEL_FORMAT format = server_init.server_pixel_format;
	// info used to find changes to screen
	frame_info info;
	info.width = screen_width;
	info.height = screen_height;
	info.bytes_per_pixel = format.bits_per_pixel/8;
	info.tiles_per_line = info.width/16 + (info.width%16 ? 1 : 0);
	info.tile_lines = info.height/16 + (info.height%16 ? 1 : 0);
	info.tiles = new bool[info.tiles_per_line*info.tile_lines];
	// port to send input events to
	port_id input_port = find_port("VNC Input port");
	if ( input_port < 0 )
		printf("Error connecting to input server add-on\n");
	else
		printf("Input server add-on port id: %ld\n", input_port );
	// keep track of which mouse buttons are pressed
	CARD8	mouse_buttons=0;
	// accepted encodings
	uint32	approved_encodings = 0;
	// frames since last full update
	int32	frames_since_full_update = 0;
	// some data for stats
	int32	total_raw_tiles = 0;
	int32	total_rre_tiles = 0;
	int32	total_frames = 0;
	bigtime_t	started_at = system_time();
	// keymap
	key_map	* keymap_keys;
	char 	* keymap_chars;
	get_key_map( &keymap_keys, &keymap_chars );
	// modifiers
	int32	curr_modifiers=0;
	// window list data
	window_list_t window_list;
	window_list.update_list = create_sem(0,"window list update request");
	window_list.list = new BList;
	window_list.updater_thread = spawn_thread(
		window_list_updater,
		"Window list updater",
		B_NORMAL_PRIORITY,
		&window_list
	);
	resume_thread( window_list.updater_thread );
	
	// Init complete, start message loop
	while ( true )
	{
		CARD8 message_type=255;
		
		if ( endpoint_read( endpoint, &message_type, sizeof(message_type) ) <= 0 )
			break;
		
		switch ( message_type )
		{
			case 0:
			{ // SetPixelFormat
				printf("SetPixelFormat\n");
				
				endpoint_read( endpoint, &format, 3 ); // read padding
				
				// read format
				endpoint_read( endpoint, &format, sizeof(format) );
				
				printf("  Bits per pixel: %d\n", (int)format.bits_per_pixel );
				printf("  Depth:          %d\n", (int)format.depth );
				printf("  Big endian:     %d\n", (int)format.big_endian );
				printf("  True color:     %d\n", (int)format.true_color );
				printf("  Red max:        %d\n", B_BENDIAN_TO_HOST_INT16(format.red_max) );
				printf("  Green max:      %d\n", B_BENDIAN_TO_HOST_INT16(format.green_max) );
				printf("  Blue max:       %d\n", B_BENDIAN_TO_HOST_INT16(format.blue_max) );
				printf("  Red shift:      %d\n", format.red_shift );
				printf("  Green shift:    %d\n", format.green_shift );
				printf("  Blue shift:     %d\n", format.blue_shift );
			}	break;
			
			case 2:
			{  // set encodings
				printf("Set encodings\n");
				CARD16 num_encodings;
				
				endpoint_read( endpoint, &num_encodings, 1 ); // skip pad
				
				// read number of encodings
				endpoint_read( endpoint, &num_encodings, sizeof(num_encodings) );
				num_encodings = B_BENDIAN_TO_HOST_INT16( num_encodings );
				printf("  %d encoding schemes: ", num_encodings);
				
				for ( int c=0; c<num_encodings; c++ )
				{
					CARD32 encoding;
					endpoint_read( endpoint, &encoding, sizeof(encoding) );
					encoding = B_BENDIAN_TO_HOST_INT32(encoding);
					
					approved_encodings |= 1 << encoding;
					switch ( encoding )
					{
						case 0: printf("raw, "); break;
						case 1: printf("copyrect, "); break;
						case 2: printf("RRE, "); break;
						case 4: printf("CoRRE, "); break;
						case 5: printf("hextile, "); break;
						
						default:
							printf("%ld, ", encoding );
					}
				}
				printf("\n");
			}	break;
			
			case 3:
			{ // frame buffer update request
				//#define DISPLAY_DEBUG
				
				framebuffer_update_request_t request;
				request.message_type = 3;
				total_frames++;
				
				endpoint_read( endpoint, &request.incremential, sizeof(request)-1 );
				
				request.x = B_BENDIAN_TO_HOST_INT16( request.x );
				request.y = B_BENDIAN_TO_HOST_INT16( request.y );
				request.width = B_BENDIAN_TO_HOST_INT16( request.width );
				request.height = B_BENDIAN_TO_HOST_INT16( request.height );
				
				// create new frame
				int32 bufsize = screen_width*screen_height*(format.bits_per_pixel/8);
				uint8 * curr_buffer = new uint8[bufsize];
				BPoint mouse_pos;
				
				#ifdef DISPLAY_DEBUG
				printf("Bits per pixel, bufsize %d/%ld\n", format.bits_per_pixel, bufsize );
				#endif
				int32 mouse_x = (int32)mouse_pos.x;
				int32 mouse_y = (int32)mouse_pos.y;
				
				// decide what parts of the screen to update
				BRegion region;
				
				if ( !last_buffer || (frames_since_full_update >= data->frames_between_full_update) )
				{ // do full update
					#ifdef DISPLAY_DEBUG
					printf("Full update\n");
					#endif
					frames_since_full_update = 0;
					region.Include( BRect(0,0,screen_width,screen_height) );
					
					if ( data->frames_between_full_update > 0 )
					{ // clear list of window frames, and update it
						release_sem( window_list.update_list );
						//update_window_frames( &window_frames );
					}
				} else
				{ // partial update, only update around mouse pointer (and current window, eventually)
					#ifdef DISPLAY_DEBUG
					printf("Partial update\n");
					#endif
					frames_since_full_update++;
					
					// copy last frame data
					memcpy( curr_buffer, last_buffer, bufsize );
					
					// make rect around mouse, 100x100
					BRect rect(mouse_pos,mouse_pos);
					rect.InsetBy(-50,-50);
					rect = rect & BRect(0,0,screen_width,screen_height); // stay on screen
					region.Include( rect );
					
					window_list.lock.Lock();
					// add any rects under mouse pointer
					for ( int c=0; c<window_list.list->CountItems(); c++ )
					{
						BRect * frame = (BRect*)window_list.list->ItemAt(c);
						if ( *frame != BRect(0,0,screen_width,screen_height) && frame->Contains( mouse_pos ) )
						{
							region.Include( *frame & BRect(0,0,screen_width,screen_height) );
							#ifdef DISPLAY_DEBUG
							frame->PrintToStream();
							#endif
						}
					}
					window_list.lock.Unlock();
					
					if ( region.Frame() == rect )
					{ // enlarge mouse pointer area if no other windows are in region
						rect.InsetBy(-50,-50);
						region.Include( rect & BRect(0,0,screen_width,screen_height) );
					}
				}
				// done setting update region
				
				// do screenshot
				#ifdef DISPLAY_DEBUG
				printf("Screenshot bounds: %.0fx%.0f\n", region.Frame().Width(), region.Frame().Height() );
				#endif
				g_fba_window->Lock();
				for ( int c=0; c<region.CountRects(); c++ )
					g_fba_window->DoScreenshot(curr_buffer, region.RectAt(c), &format );
				mouse_pos = g_fba_window->GetMousePos();
				g_fba_window->Unlock();
				
				// draw pointer
				#ifdef DISPLAY_DEBUG
				printf("Drawing pointer\n");
				#endif
				for ( int y=0; y<POINTER_HEIGHT; y++ )
					for ( int x=0; x<POINTER_WIDTH; x++ )
					{
						if ( mouse_pointer[y][x].solid && x+mouse_x < screen_width && y+mouse_y < screen_height )
						{
							int32 offset = (mouse_y+y)*info.bytes_per_pixel*info.width + info.bytes_per_pixel*(mouse_x+x);
							memcpy( &curr_buffer[offset], &mouse_pointer[y][x].data, info.bytes_per_pixel );
						}
					}
					
				// count changed tiles
				int num_tiles = info.tiles_per_line * info.tile_lines;
				int changed_tiles = 0;
				
				if ( last_buffer && request.incremential )
				{ // find differences
					info.next_buffer = curr_buffer;
					info.prev_buffer = last_buffer;
					
					memset(info.tiles,0,info.tiles_per_line*info.tile_lines);
					update_frame_info( &info );
					
					for ( int c=0; c<num_tiles; c++ )
						if ( info.tiles[c] )
							changed_tiles++;
					
					if ( changed_tiles == 0 )
					{ // whoa! nothing's changed! fake a change..
						changed_tiles = 1;
						info.tiles[0] = true;
					}
				} else
				{ // full update, set all tiles to true
					for ( int c=0; c<num_tiles; c++ )
						info.tiles[c] = true;
					
					changed_tiles = num_tiles;
				}
				
				// buffer entire update and send it in a single chunk (otherwise the net_server chokes, more or less)
				SmartBuffer send_buffer(bufsize*2);
				
				// send data
				framebuffer_update_t update;
				update.message_type = 0; // FRAMEBUFFER UPDATE MESSAGE
				update.num_rect = B_HOST_TO_BENDIAN_INT16( changed_tiles );
				
				send_buffer.Add( &update, sizeof(update) );
				
				for ( int tile_y=0; tile_y<info.tile_lines; tile_y++ )
				{
					for ( int tile_x=0; tile_x<info.tiles_per_line; tile_x++ )
					{ // for each tile
						if ( info.tiles[tile_y*info.tiles_per_line+tile_x] )
						{
							// calc some offsets
							uint32 top_left_offset = (tile_y*16) * info.width * info.bytes_per_pixel;
							top_left_offset += tile_x * 16 * info.bytes_per_pixel;
							// width of lines
							int32 pixels_to_copy = 16;
							if ( (tile_x+1) * 16 > info.width )
								pixels_to_copy = info.width % 16;
							int32 to_copy = pixels_to_copy * info.bytes_per_pixel;
							// no of lines to send
							int32 lines = 16;
							if ( (tile_y+1) * 16 > info.height )
								lines = info.height%16;
								
							// set encoding to use
							
							// check if solid color
							bool is_solid = true;
							char color_data[4];
								
							memcpy(color_data, &curr_buffer[top_left_offset], info.bytes_per_pixel);
								
							for ( int y=0; y<lines && is_solid; y++ )
							{
								int32 offset = top_left_offset + y*(info.width * info.bytes_per_pixel);
								
								for ( int x=0; x<pixels_to_copy; x++ )
								{
									if ( memcmp( 
										color_data, 
										&curr_buffer[offset+x*info.bytes_per_pixel], 
										info.bytes_per_pixel )
									)
									{
										is_solid = false;
										break;
									}
								}
							}
							
							// use RRE if solid, else RAW
							CARD32 encoding;
							if ( is_solid && (approved_encodings & (1 << ENCODING_RRE)) )
								encoding = ENCODING_RRE;
							else
								encoding = ENCODING_RAW;
							
							// send rect info
							rectangle_t rect;
							rect.x = B_HOST_TO_BENDIAN_INT16( tile_x*16 );
							rect.y = B_HOST_TO_BENDIAN_INT16( tile_y*16 );
							rect.height = B_HOST_TO_BENDIAN_INT16( lines );
							rect.width = B_HOST_TO_BENDIAN_INT16( pixels_to_copy );
							rect.encoding = B_HOST_TO_BENDIAN_INT32(encoding);
							
							send_buffer.Add( &rect, sizeof(rect) );
							
							switch ( encoding )
							{
								case ENCODING_RAW:
								{ // RAW
									total_raw_tiles++;
									
									for ( int l=0; l<lines; l++ )
									{ // send each line
										uint32 offset = top_left_offset + l*(info.width * info.bytes_per_pixel);
										send_buffer.Add( &curr_buffer[offset], to_copy );
									}
								}	break;
								
								case ENCODING_RRE:
								{ // RRE
									total_rre_tiles++;
									
									int32 num_sub_rects = 0;
									
									send_buffer.Add( &num_sub_rects, sizeof(num_sub_rects) );
									send_buffer.Add( &color_data, info.bytes_per_pixel );
								}	break;
								
								default:
									printf("Error: Trying to send using unknown encoding\n");
									delete endpoint;
									return 1;
							}
						}
					}
				}
					
				// send buffered data
				endpoint_write( endpoint, send_buffer.Buffer(), send_buffer.Size() );
				
				// free old buffer
				if ( last_buffer )
					delete [] last_buffer;
				last_buffer = curr_buffer;
			}	break;
			
			case 4:
			{ // key event message
				//#define KEYBOARD_DEBUG
				
				key_event_t event;
				
				endpoint_read( endpoint, &event.down, sizeof(event)-1 );
				
				event.key = B_BENDIAN_TO_HOST_INT32( event.key );
				
				BMessage msg;
				
				if ( event.down )
					msg.what = B_KEY_DOWN;
				else
					msg.what = B_KEY_UP;
				
				char string[4] = {0,0,0,0};
				
				switch ( event.key )
				{ // watch for special keys
					case 0xffe9: case 0xffea:
					{ // COMMAND
						if ( event.down )
							curr_modifiers |= B_COMMAND_KEY | B_LEFT_COMMAND_KEY;
						else
							curr_modifiers &= (0xffffffff - (B_COMMAND_KEY | B_LEFT_COMMAND_KEY));
					}	break;
					case 0xffe3: case 0xffe4:
					{ // CONTROL
						if ( event.down )
							curr_modifiers |= B_CONTROL_KEY | B_LEFT_CONTROL_KEY;
						else
							curr_modifiers &= (0xffffffff - (B_CONTROL_KEY | B_LEFT_CONTROL_KEY));
					}	break;
					case 0xffe1: case 0xffe2:
					{ // SHIFT
						if ( event.down )
							curr_modifiers |= B_SHIFT_KEY | B_LEFT_SHIFT_KEY;
						else
							curr_modifiers &= (0xffffffff - (B_SHIFT_KEY  | B_LEFT_SHIFT_KEY));
					}	break;
					case 0xff08:
					{ // B_BACKSPACE
						msg.AddInt32("key", 0x1e);
						string[0] = B_BACKSPACE;
					}	break;
					case 0xff09:
					{ // B_TAB
						msg.AddInt32("key", 0x26);
						string[0] = B_TAB;
					}	break;
					case 0xff0d:
					{ // B_RETURN
						msg.AddInt32("key", 0x47);
						string[0] = B_RETURN;
					}	break;
					case 0xff1b:
					{ // B_ESCAPE
						msg.AddInt32("key", 0x01);
						string[0] = B_ESCAPE;
					}	break;
					case 0xff63:
					{ // B_INSERT
						msg.AddInt32("key", 0x1f);
						string[0] = B_INSERT;
					}	break;
					case 0xffff:
					{	// B_DELETE;
						msg.AddInt32("key", 0x34);
						string[0] = B_DELETE;
					}	break;
					case 0xff50:
					{ // B_HOME
						msg.AddInt32("key", 0x20);
						string[0] = B_HOME;
					}	break;
					case 0xff57:
					{ // B_END
						msg.AddInt32("key", 0x35);
						string[0] = B_END;
					}	break;
					case 0xff55:
					{ // B_PAGE_UP
						msg.AddInt32("key", 0x21);
						string[0] = B_PAGE_UP;
					}	break;
					case 0xff56:
					{ // B_PAGE_DOWN
						msg.AddInt32("key", 0x36);
						string[0] = B_PAGE_DOWN;
					}	break;
					case 0xff51:
					{ // B_LEFT_ARROW
						msg.AddInt32("key", 0x61);
						string[0] = B_LEFT_ARROW;
					}	break;
					case 0xff52:
					{ // B_UP_ARROW
						msg.AddInt32("key", 0x57);
						string[0] = B_UP_ARROW;
					}	break;
					case 0xff53:
					{ // B_RIGHT_ARROW
						msg.AddInt32("key", 0x63);
						string[0] = B_RIGHT_ARROW;
					}	break;
					case 0xff54:
					{ // B_DOWN_ARROW
						msg.AddInt32("key", 0x62);
						string[0] = B_DOWN_ARROW;
					}	break;
					case 0xffbe: case 0xffbf: case 0xffc0: case 0xffc1:
					case 0xffc2: case 0xffc3: case 0xffc4: case 0xffc5:
					case 0xffc6: case 0xffc7: case 0xffc8: case 0xffc9:
					{ // B_FUNCTION_KEY
							string[0] = B_FUNCTION_KEY;
							msg.AddInt32("key", B_F1_KEY + event.key - 0xffbe);
					}	break;
					
					default:
						if 	( find_key_in_keymap(
								event.key,
								keymap_keys->normal_map,
								keymap_chars,
								&msg,
								curr_modifiers)
						)	break;
						
						if 	( find_key_in_keymap(
								event.key,
								keymap_keys->shift_map,
								keymap_chars,
								&msg,
								curr_modifiers)
						)	break;
						
						break;
				}
				
				if ( string[0] )
				{ // special key, add 'byte' and 'bytes' fields to message
					msg.AddString("bytes", string);
					for ( int c=0; string[c]; c++ )
						msg.AddInt8( "byte", string[c] );
					msg.AddInt32("modifiers", curr_modifiers );
				}
				
				#ifdef KEYBOARD_DEBUG
				msg.PrintToStream();
				#endif
				
				if ( msg.FlattenedSize() > 20 )
				{ // flatten and send if 
					char * buffer = new char[msg.FlattenedSize()];
					msg.Flatten( buffer, msg.FlattenedSize() );
					if ( write_port(input_port,123,buffer,msg.FlattenedSize()) != B_OK )
						printf("Error writing to input add-on port\n");
					delete [] buffer;
				}
				
			}	break;
			
			case 5:
			{ // pointer event message
				//#define POINTER_DEBUG
				pointer_event_t event;
				
				endpoint_read( endpoint, &event.buttons, sizeof(event)-1 );
				
				event.x = B_BENDIAN_TO_HOST_INT16( event.x );
				event.y = B_BENDIAN_TO_HOST_INT16( event.y );
				
				// convert position to float ( 0.0 - 1.0 )
				float fx = (float)event.x / (float)info.width;
				float fy = (float)event.y / (float)info.height;
				
				BMessage msg(B_MOUSE_MOVED);
				msg.AddFloat("x", fx);
				msg.AddFloat("y", fy);
				msg.AddInt32("buttons", event.buttons);
				msg.AddInt32("modifiers", curr_modifiers);
				
				// flatten and send message
				char * msgbuffer = new char[msg.FlattenedSize()];
				msg.Flatten( msgbuffer, msg.FlattenedSize() );
				if ( write_port(input_port,123,msgbuffer,msg.FlattenedSize()) != B_OK )
					printf("Error writing to input add-on port\n");
				delete [] msgbuffer;
				#ifdef POINTER_DEBUG
				msg.PrintToStream();
				#endif
				
				if ( event.buttons != mouse_buttons )
				{ // button(s) pressed or released
					for ( int c=0; c<8;c++ )
					{
						if ( (mouse_buttons & (1 << c)) != (event.buttons & (1 << c)) )
						{ // button c pressed/released
							if ( (event.buttons & (1 << c)) )
								msg.what = B_MOUSE_DOWN;
							else
								msg.what = B_MOUSE_UP;
							msg.AddInt32("buttons", event.buttons);
							
							// flatten and send message
							msgbuffer = new char[msg.FlattenedSize()];
							msg.Flatten(msgbuffer,msg.FlattenedSize());
							if ( write_port(input_port,123,msgbuffer,msg.FlattenedSize()) != B_OK )
								printf("Error writing to input add-on port\n");
							delete [] msgbuffer;
						}
					}
					
					mouse_buttons = event.buttons;
				}
				
				#ifdef POINTER_DEBUG
				printf("End mouse event\n");
				#endif
			}	break;
			
			default:
				printf("Unknown message %d\n", (int)message_type );
				endpoint->Close();
				be_app->PostMessage( B_QUIT_REQUESTED );
				break;
		}
	}
	
	printf("Client disconnecting\n");
	
	printf("  Shutting down window lister\n");
	status_t res=0;
	delete_sem( window_list.update_list );
	wait_for_thread( window_list.updater_thread, &res );
	
	bigtime_t ended_at = system_time();
	
	float time_elapsed = (ended_at - started_at) / 1000000.0;
	float fps = total_frames / time_elapsed;
	float bps_out = g_total_data_sent / time_elapsed;
	float bps_in = g_total_data_read / time_elapsed;
	
	printf("Stats:\n");
	printf("  Session lasted     %Ld us\n", ended_at - started_at );
	printf("  Frames sent:       %ld\n", total_frames );
	printf("  Frames/sec:        %.02f\n", fps);
	printf("  Total raw tiles:   %ld\n", total_raw_tiles);
	printf("  Total RRE tiles:   %ld\n", total_rre_tiles);
	printf("  Bytes/sec sent:    %.02f\n", bps_out);
	printf("  Bytes/sec read:    %.02f\n", bps_in);
	delete endpoint;
	
	return 0;
};

/*
	Listens for connections and starts client threads
*/
int32 listener_thread( void * _data )
{
	client_data_t * data = (client_data_t*)_data;
	
	BNetEndpoint listener;
	listener.Bind( 5900 );
	
	listener.Listen();
	
	while ( true )
	{
		BNetEndpoint * client = listener.Accept();
		
		if ( !client )
			break;
		
		data->endpoint = client;
		
		thread_id connection = spawn_thread(
			client_thread,
			"client",
			B_NORMAL_PRIORITY,
			data
		);
		resume_thread( connection );
	}
	
	return 0;
}

class VNCserver_app : public BApplication
{
	public:
		VNCserver_app()
		:	BApplication("application/x-vnd.m_eiman.vncserver")
		{
		};
		
		virtual ~VNCserver_app()
		{
		};
		
		virtual void MessageReceived( BMessage * msg )
		{
			BApplication::MessageReceived(msg);
		};
};

int main( int numarg, char ** argv )
{
	VNCserver_app app;
	
	client_data_t client_data;
	client_data.frames_between_full_update = 10;
	client_data.compress_bitdepth = false;
	
	for ( int c=0; c<numarg; c++ )
	{ // display help
		if ( strcasecmp("-h", argv[c]) == 0 || strcasecmp("-?", argv[c]) == 0 || strcasecmp("--help", argv[c]) == 0 )
		{
			printf("Usage: vncserver [options]\n");
			printf("  Options:\n");
			printf("    -h        Display this help\n");
			printf("    -c        Compress bitdepth (use 8 bits/pixel instead of 16 or 32)\n");
			printf("    -f n      Do full update every n'th frame (default n=10)\n");
			exit(0);
		}
	}
	
	for ( int c=0; c<numarg; c++ )
	{ // compress
		if ( strcasecmp("-c", argv[c]) == 0 )
		{
			client_data.compress_bitdepth = true;
		}
	}
	
	for ( int c=0; c<numarg; c++ )
	{ // frames between full update
		if ( strcasecmp("-f", argv[c]) == 0 )
		{
			if ( c != numarg-1 )
			{
				client_data.frames_between_full_update = atoi(argv[c+1]);
			}
		}
	}
	
	printf("Settings:\n");
	printf("  Full update every %d frames\n", client_data.frames_between_full_update);
	printf("  Bitdepth compression is %s\n", client_data.compress_bitdepth ? "on" : "off");
	
	g_fba_window = new FBAWindow();
	
	thread_id thread = spawn_thread(
		listener_thread,
		"listener",
		B_NORMAL_PRIORITY,
		&client_data
	);
	resume_thread( thread );
	
	app.Run();
	
	printf("Data read: %ld\n", g_total_data_read );
	printf("Data sent: %ld\n", g_total_data_sent );
	printf("Done\n" );
	
	return 0;
}
